/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

/*
begin_html
<p>
 See also:
 <ul>
 </ul>
 <b>Up:</b> <a href="index.html">Index</a>
</p>
end_html
*/

#ifdef _WIN32
  #include <windows.h>
#else
  #include <stdlib.h>
  #define BOOL int
  #define TRUE 1
  #define FALSE 0
#endif

#include <iostream>
#include <fstream>
#include <vector>

#include "qub_kalman.h"
#include "ublas_matrixutil.h"

using std::cout;
using std::cerr;
using std::endl;



class qub_kalman_storage
{
public:
	int Np, Nm;
	pointy_col_vector<double> X, Z;
	pointy_matrix<double> P, A, Q, H, R;
	pointy_col_vector<double> Xprior, ZmHx;
	pointy_matrix<double> Pprior, PpAt, PpHt, HPpHt, HPpHtRi, K, Ip, KH, ImKH;

	qub_kalman_storage(int Nprocess, int Nmeasure)
		: Np(Nprocess), Nm(Nmeasure), X(Nprocess), Z(Nmeasure),
		P(Nprocess, Nprocess), A(Nprocess, Nprocess), Q(Nprocess, Nprocess),
		H(Nmeasure, Nprocess), R(Nmeasure, Nmeasure),
		Xprior(Nprocess), ZmHx(Nmeasure), Pprior(Nprocess, Nprocess),
		PpAt(Nprocess, Nprocess), PpHt(Nprocess, Nmeasure),
		HPpHt(Nmeasure, Nmeasure), HPpHtRi(Nmeasure, Nmeasure),
		K(Nprocess, Nmeasure), Ip(Nprocess, Nprocess),
		KH(Nprocess, Nprocess), ImKH(Nprocess, Nprocess)
	{
		fill_identity(Ip.m);
		fill_identity(A.m);
	}

	void next(double *z)
	{
		noalias(Xprior.m) = prod(A.m, X.m);          // x- = A*x;
		noalias(PpAt.m)   = prod(P.m, trans(A.m));
		noalias(Pprior.m) = prod(A.m, PpAt.m);     // p- = A*P*At
		Pprior.m         += Q.m;                   //      + Q;

		noalias(PpHt.m)   = prod(Pprior.m, trans(H.m));
		noalias(HPpHt.m)  = prod(H.m, PpHt.m);
		HPpHt.m          += R.m;
		if ( Nm > 1 )
			sv_invm(HPpHt.pp, Nm, HPpHtRi.pp);         // inv(H*P-*Ht + R)
		else
			HPpHtRi.p[0] = 1.0 / HPpHt.p[0];
		noalias(K.m)    = prod(PpHt.m, HPpHtRi.m); // K = P-*Ht * inv(HP-Ht + R);

		noalias(ZmHx.m) = prod(H.m, Xprior.m);
		for (int i=0; i<Nm; ++i)
			ZmHx.p[i] = z[i] - ZmHx.p[i];          // (Z - H*Xp)
		noalias(X.m)    = prod(K.m, ZmHx.m);
		for (int i=0; i<Np; ++i)
			X.p[i]   += Xprior.p[i];               // X = Xp + K(Z - HXp);

		noalias(KH.m)   = prod(K.m, H.m);
		ImKH.m          = Ip.m - KH.m;
		noalias(P.m)    = prod(ImKH.m, Pprior.m);  // P = (I - KH)Pp;

		noalias(Z.m)    = prod(H.m, X.m);          // Z = HX;
	}
	void reset()
	{
		P.m = Q.m;
	}
};

extern "C" MAXILL_API qub_kalman_filter * qub_kalman_filter_create(int Nprocess, int Nmeasure)
{
	qub_kalman_filter *f = new qub_kalman_filter;
	qub_kalman_storage *s = new qub_kalman_storage(Nprocess, Nmeasure);
	f->storage = s;
	f->Nprocess = Nprocess;
	f->Nmeasure = Nmeasure;
	f->X = s->X.p;
	f->Z = s->Z.p;
	f->P = s->P.pp;
	f->A = s->A.pp;
	f->Q = s->Q.pp;
	f->H = s->H.pp;
	f->R = s->R.pp;
	return f;
}

extern "C" MAXILL_API void qub_kalman_filter_free(qub_kalman_filter *filter)
{
	if ( filter )
		delete (qub_kalman_storage*) filter->storage;
	delete filter;
}

extern "C" MAXILL_API void qub_kalman_filter_next(qub_kalman_filter *filter, double *z)
{
	((qub_kalman_storage*) (filter->storage))->next(z);
}

extern "C" MAXILL_API void qub_kalman_filter_reset(qub_kalman_filter *filter)
{
	((qub_kalman_storage*) (filter->storage))->reset();
}
